/*

mtf - a Microsoft Tape Format reader (and future writer?)
Copyright (C) 1999  D. Alan Stewart, Layton Graphics, Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

Contact the author at:

D. Alan Stewart
Layton Graphics, Inc.
155 Woolco Dr.
Marietta, GA 30062
astewart@layton-graphics.com

**
**	mtf.c
**
**	This is the main source code file for the progam mtf. It is a bare-bones
**	read for Microsoft Tape Format tapes. Many things unsupported!
**
**	Version history:
**
**	0.1	D. Alan Stewart, Layton Graphics, Inc. 3/22/1999
**
*/

#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/mtio.h>
#include <regex.h>
#include <grp.h>
#include <pwd.h>
#include "mtf.h"


char outPath[MAXPATHLEN + 1], curPath[MAXPATHLEN + 1], device[MAXPATHLEN + 1];
int mtfd;
UINT8 verbose, debug, list, forcecase;
UINT16 setNum, matchCnt;
UINT32 minFree;
regex_t match[MAX_PATTERN];
gid_t group;
uid_t owner;


int main(int, char*[]);
INT16 parseArgs(int, char*[]);
INT16 whichSet(char*);
INT16 whichDevice(char*);
INT16 setPath(char*);
INT16 setOwner(char*);
INT16 setGroup(char*);
INT16 setCase(char*);
INT16 setMinFree(char*);
INT16 getPatterns(int, char*[], int);
void usage(void);

int main(int argc, char *argv[])
{
	char *ptr, buffer[MAXPATHLEN + 1];
	int i;
	INT16 result;
	struct stat sbuf;
	struct mtop op;
	struct mtget get;

	if (parseArgs(argc, argv) != 0)
	{
		return(-1);
	}

	if (verbose > 0) fprintf(stdout, "Verbose mode selected.\n");
	if (verbose > 1) fprintf(stdout, "Very verbose mode selected.\n");
	if ((verbose > 0) && (list > 0)) fprintf(stdout, "List mode selected.\n");

	if (debug > 0)
	{
		verbose = 2;
		fprintf(stdout, "Debug mode selected.\n");
	}

	if (strlen(outPath) == 0)
	{
		strcpy(outPath, ".");
	}
	else if (outPath[0] != '/')
	{
		getcwd(curPath, MAXPATHLEN);
		strcpy(buffer, outPath);
		sprintf(outPath, "%s/%s", curPath, buffer);

		if (outPath[strlen(outPath) - 1] == '/')
			outPath[strlen(outPath) - 1] = '\0';
	}

	if (stat(outPath, &sbuf) != 0)
	{
		fprintf(stderr, "Error testing for status of %s!\n", outPath);
		return(-1);
	}

	if ((sbuf.st_mode & S_IFDIR) == 0)
	{
		fprintf(stderr, "%s is not a directory!\n", outPath);
		return(-1);
	}

	strcpy(curPath, "");

	if (verbose > 0)
		fprintf(stdout, "Output will be written to %s.\n", outPath);

	if (strlen(device) == 0)
	{
		ptr = getenv("TAPE");
		if (ptr != NULL)
		{
			ptr = strrchr(ptr, '=');
			if (ptr != NULL)
			{
				while ((*ptr == '=') || (*ptr == ' '))
					ptr += 1;

				strcpy(device, ptr);
			}
		}
	}

	if (strlen(device) == 0)
		strcpy(device, "/dev/tape");

	if (verbose > 0) fprintf(stdout, "Device is set to %s.\n", device);

	mtfd = open(device, O_RDONLY);
	if (mtfd < 0)
	{
		fprintf(stderr, "Error opening %s for reading!\n", device);
		return(-1);
	}

	if (openMedia() != 0)
	{
		fprintf(stderr, "Error opening tape!\n");
		goto error;
	}

	if (setNum > 1)
	{
		op.mt_op = MTFSF;
		op.mt_count = (setNum - 1) * 2;

		if (verbose > 0)
			fprintf(stdout, "Forwarding tape to data set #%u...\n", setNum);

		if (ioctl(mtfd, MTIOCTOP, &op) != 0)
		{
			fprintf(stderr, "Error forwarding tape!\n");
			close(mtfd);
			goto error;
		}

		if (debug > 0)
		{
			op.mt_op = MTNOP;
			op.mt_count = 0;

			if (ioctl(mtfd, MTIOCTOP, &op) != 0)
			{
				fprintf(stderr, "Error returned by MTIOCTOP!\n");
				goto error;
			}

			if (ioctl(mtfd, MTIOCGET, &get) != 0)
			{
				fprintf(stderr, "Error returned by MTIOCGET!\n");
				goto error;
			}

			printf("tape file no. %u\n", get.mt_fileno);
			printf("tape block no. %u\n", get.mt_blkno);
		}

		if (readNextBlock(0) != 0)
		{
			fprintf(stderr, "Error reading first block of data set!\n");
			goto error;
		}
	}

next:

	if (readDataSet() != 0)
	{
		fprintf(stderr, "Error reading data set!\n");
		goto error;
	}

	result = readEndOfDataSet();
	if (result < 0)
	{
		fprintf(stderr, "Error reading to end of data set!\n");
		goto error;
	}

	if ((result == 0) && (setNum == 0))
		goto next;

	if (verbose > 0) fprintf(stdout, "Successful read of archive!\n");
	
	close(mtfd);

	return(0);

error:

	for (i = 0; i < matchCnt; i += 1)
	{
		regfree(&match[i]);
	}

	dump("errorblock.dmp");
	close(mtfd);

	return(-1);
}


INT16 parseArgs(int argc, char *argv[])
{
	int i, j;
	char *ptr;

	verbose = 0;
	debug = 0;
	list = 0;
	setNum = 0;
	strcpy(device, "");
	strcpy(outPath, "");
	matchCnt = 0;
	minFree = 0;
	owner = -1;
	group = -1;
	forcecase = CASE_SENSITIVE;

	i = 1;

	while (i < argc)
	{
		if (argv[i][0] == '-')
		{
			ptr = &argv[i][1];
			j = i;

			while ((*ptr != '\0') && (i == j))
			{
				if (strcmp(ptr, "s") == 0)
				{
					i += 1;

					if (whichSet(argv[i]) != 0)
						return(-1);

					i += 1;
				}
				else if (strcmp(ptr, "d") == 0)
				{
					i += 1;

					if (whichDevice(argv[i]) != 0)
						return(-1);

					i += 1;
				}
				else if (strcmp(ptr, "o") == 0)
				{
					i += 1;

					if (setPath(argv[i]) != 0)
						return(-1);

					i += 1;
				}
				else if (strcmp(ptr, "u") == 0)
				{
					i += 1;

					if (setOwner(argv[i]) != 0)
						return(-1);

					i += 1;
				}
				else if (strcmp(ptr, "g") == 0)
				{
					i += 1;

					if (setGroup(argv[i]) != 0)
						return(-1);

					i += 1;
				}
				else if (strcmp(ptr, "c") == 0)
				{
					i += 1;

					if (setCase(argv[i]) != 0)
						return(-1);

					i += 1;
				}
				else if (strcmp(ptr, "f") == 0)
				{
					i += 1;

					if (setMinFree(argv[i]) != 0)
						return(-1);

					i += 1;
				}
				else if (*ptr == 'v')
				{
					verbose = 1;
					ptr += 1;
				}
				else if (*ptr == 'V')
				{
					verbose = 2;
					ptr += 1;
				}
				else if (*ptr == 'D')
				{
					debug = 1;
					ptr += 1;
				}
				else if (*ptr == 'l')
				{
					list = 1;
					ptr += 1;
				}
				else
				{
					fprintf(stderr, "Unrecognized arguemnt!\n");
					usage();
					return(-1);
				}

				if (*ptr == '\0')
					i += 1;
			}
		}
		else if (argv[i][0] != '-')
		{
			if (getPatterns(argc, argv, i) != 0)
				return(-1);
			else
				return(0);
		}
		else
		{
			fprintf(stderr, "Unrecognized arguemnt!\n");
			usage();
			return(-1);
		}
	}

	return(0);
}


INT16 whichSet(char *argv)
{
	long test;

	if (strlen(argv) == 0)
	{
		fprintf(stderr, "No value given for set number (-s)!\n");
		usage();
		return(-1);
	}

	if (strcmp(argv, "*") == 0)
	{
		setNum = 0;
	}
	else
	{
		test = atoi(argv);

		if ((test < 1) || (test > 65535))
		{
			fprintf(stderr,
			        "Value given for set number (-s) is out of range!\n");
			usage();
			return(-1);
		}

		setNum = test;
	}

	return(0);
}


INT16 whichDevice(char *argv)
{
	if (strlen(argv) == 0)
	{
		fprintf(stderr, "No value given for device (-d)!\n");
		usage();
		return(-1);
	}

	strcpy(device, argv);

	return(0);
}


INT16 setPath(char *argv)
{
	if (strlen(argv) == 0)
	{
		fprintf(stderr, "No value given for output path (-o)!\n");
		usage();
		return(-1);
	}

	strcpy(outPath, argv);

	return(0);
} 


INT16 setOwner(char *argv)
{
	struct passwd *pbuf;

	if (strlen(argv) == 0)
	{
		fprintf(stderr, "No value given for owner (-u)!\n");
		usage();
		return(-1);
	}

	pbuf = getpwnam(argv);
	if (pbuf == NULL)
	{
		fprintf(stderr, "Unable to retrieve user id (-u) of %s!\n", argv);
		usage();
		return(-1);
	}

	owner = pbuf->pw_uid;

	return(0);
} 


INT16 setGroup(char *argv)
{
	struct group *gbuf;

	if (strlen(argv) == 0)
	{
		fprintf(stderr, "No value given for group (-g)!\n");
		usage();
		return(-1);
	}

	gbuf = getgrnam(argv);
	if (gbuf == NULL)
	{
		fprintf(stderr, "Unable to retrieve group id (-g) of %s!\n", argv);
		usage();
		return(-1);
	}

	group = gbuf->gr_gid;

	return(0);
} 


INT16 setCase(char *argv)
{
	if (strlen(argv) == 0)
	{
		fprintf(stderr, "No value given for forcing case (-c)!\n");
		usage();
		return(-1);
	}

	strlwr(argv);

	if (strcmp(argv, "lower") == 0)
		forcecase = CASE_LOWER;
	else if (strcmp(argv, "upper") == 0)
		forcecase = CASE_UPPER;
	else
	{
		fprintf(stderr, "Invalid value given for forcing case (-c) - \"%s\"!\n", argv);
		usage();
		return(-1);
	}

	return(0);
} 


INT16 setMinFree(char *argv)
{
	char *ptr;

	if (strlen(argv) == 0)
	{
		fprintf(stderr, "No value given for minimum free space (-f)!\n");
		usage();
		return(-1);
	}

	strlwr(argv);

	ptr = argv;
	while ((*ptr >= '0') && (*ptr <= '9'))
		ptr += 1;
	
	if (strlen(ptr) == 0)
		minFree = (UINT32) atol(argv);
	else if (strcmp(ptr, "k") == 0)
		minFree = (UINT32) atol(argv) * 1024;
	else if (strcmp(ptr, "m") == 0)
		minFree = (UINT32) atol(argv) * 1024 * 1024;
	else
	{
		fprintf(stderr,
		        "Invalid value given for minimum free space (-f) - \"%s\"!\n",
				argv);
		usage();
		return(-1);
	}

	return(0);
} 


INT16 getPatterns(int argc, char *argv[], int start)
{
	int i;

	i = start;
	while (i < argc)
	{
		if (argv[i][0] == '-')
		{
			fprintf(stderr, "Error parsing pattern!\n");
			usage();
			return(-1);
		}

		if (strlen(argv[i]) >= MAX_PATTERN_LEN)
		{
			fprintf(stderr, "Pattern exceeds maximum length!\n");
			usage();
			return(-1);
		}

		if (matchCnt == MAX_PATTERN)
		{
			fprintf(stderr, "Maximum number of patterns exceeded!\n");
			usage();
			return(-1);
		}

		if (regcomp(&match[matchCnt], argv[i],
		            REG_ICASE | REG_NOSUB | REG_NEWLINE) != 0)
		{
			fprintf(stderr, "Invalid pattern - \"%s\"!\n", argv[i]);
			usage();
			return(-1);
		}

		matchCnt += 1;

		i += 1;
	}

	return(0);
}


void usage(void)
{
	fprintf(stderr, "Usage: mtf [options] [pattern(s)]\n");
	fprintf(stderr, "    -v               verbose\n");
	fprintf(stderr, "    -V               very verbose\n");
	fprintf(stderr, "    -D               debug\n");
	fprintf(stderr, "    -l               list contents\n");
	fprintf(stderr, "    -d device        device to read from\n");
	fprintf(stderr, "    -s set           number of data set to read\n");
	fprintf(stderr, "    -u user          assign owner to all files/directories written;\n");
	fprintf(stderr, "                     useful only for the root user\n");
	fprintf(stderr, "    -g group         assign group to all files/directories written;\n");
	fprintf(stderr, "                     useful only for the root user\n");
	fprintf(stderr, "    -c [lower|upper] force the case of paths\n");
	fprintf(stderr, "    -o path          root path to write files to\n");
	fprintf(stderr, "    -f bytes[K|M]    maintain minimum free space of bytes;\n");
	fprintf(stderr, "                     a K or M suffix signifies kilobytes or megabytes\n");
	fprintf(stderr, "    pattern(s)       only read paths that match one of the patterns\n");

	return;
}

